/*
  # This code is based on the PH test code from Grove for the PH Sensor Kit (E-201C-Blue). (https://files.seeedstudio.com/wiki/Grove-PH_Sensor_kit/Grove-PH_Sensor_kit.zip)
  # It is also based on the sample servo program.
  #
*/

// inclusions
// these libraries can be downloaded through the Arduino interface
#include <NeoSWSerial.h>
#include <NMEAGPS.h>
#include <Servo.h>


// ph definitions
#define SensorPin A0          //pH meter Analog output to Arduino Analog Input 0
//#define Offset 41.25746741
#define Offset 40.55746741
#define samplingInterval 20
#define printInterval 1000
#define ArrayLenth  40  //times of collection

// global variables
int bdrate = 9600;    // global baud rate value

// pH variables
int pHArray[ArrayLenth];   //Store the average value of the sensor feedback
int pHArrayIndex = 0;
int pHControlPin = 7;

// water values
int WATER_THRESHOLD = 200;
int inWater = true;
int waterPin = A1;
long waterTimer = 0;
static const int waterInterval = 100; // interval in ms between checks whether it is in water.

// GPS values
static NMEAGPS gps;
static gps_fix fix;
static const int gpsInterval = 500; // interval in ms when the gps checks for position
NeoSWSerial gpsPort(8, 9);

// Servo Values
static const int servoPin = 3;        // pin the servo attaches to.
static const long measuringTime = 120000; // time the sensor will be kept in the water in milliseconds (currently 2 minutes).
static const long noWaterTime = 10000;  // timeout of servo when it isn't in water (currently 10 seconds
static const int measuringAngle = 110;  // angle of the servo for measuring
static const int standbyAngle = 0;      // angle of the servo on standby
long servoTimer = 999999; // timer that keeps track of time past since trying to measure.
//long moveTimer = LONG_MAX;  // timer that makes sure the servo has time to catch up.
boolean measuring;  // boolean that stores whether the product is trying to measure.
Servo ourservo;   // servo object
//int servoPos;

// Wireless Communication variables
//NeoSWSerial wlCom(11, 12);
long commTimer = 500;
static const long commInterval = 1000;

// start of functional code
void setup() {
  Serial.begin(bdrate);
//  wlCom.begin(bdrate);
  pHSetup();
  waterSetup();
  gPSSetup();
  ourservo.attach(servoPin); // attaches servo to the servoPin;
}

void loop() {
  pHLoop();
  waterLoop();
  gPSLoop();
  commLoop();
  servoLoop();
}

inline void commLoop() {
  if (millis() >= commTimer + commInterval) {
  commTimer = millis();
  if (Serial.available() > 0) {
    String cmd = Serial.readString();
    cmd.trim();
    if (cmd.length() > 0) {
      measureCommand();
    }
  }
  }
}

inline void servoLoop() {
  if (measuring) {
  ourservo.write(measuringAngle);
  if (( inWater && millis() >= servoTimer + measuringTime) || ( !inWater && millis() >= servoTimer + noWaterTime)) { 
    measuring = false;
  }
  } else {
  ourservo.write(standbyAngle);
  }
}

void measureCommand() {
  measuring = true;
  servoTimer = millis();
}

inline void gPSSetup() {
  gpsPort.begin(bdrate);
}

inline void gPSLoop() {
  if (gps.available( gpsPort ))// && millis() - timer > gpsInterval)
  GPSPrintFix();
  //gpsTimer = millis();
}

inline void waterSetup() {
  pinMode(waterPin, INPUT);
}

inline void waterLoop() {
  if (millis() - waterTimer > waterInterval) {
  int waterVal = analogRead(waterPin);
  // controls the control pin
  if (waterVal > WATER_THRESHOLD && !inWater) {
    inWater = true;
    digitalWrite(pHControlPin, HIGH);
  } else if (waterVal <= WATER_THRESHOLD && inWater) {
    inWater = false;
    digitalWrite(pHControlPin, LOW);
  }
  }
}

inline void pHSetup(void)
{
  pinMode(pHControlPin, OUTPUT);
  digitalWrite(pHControlPin, LOW);
}

inline void pHLoop(void)
{
  // controls the measuring
  static unsigned long samplingTime = millis();
  static unsigned long printTime = millis();
  static float pHValue, voltage;
  if (millis() - samplingTime > samplingInterval)
  {
  pHArray[pHArrayIndex++] = analogRead(SensorPin);
  if (pHArrayIndex == ArrayLenth)pHArrayIndex = 0;
  voltage = averagearray(pHArray, ArrayLenth) * 5.0 / 1024;
  pHValue = -19.185 * voltage + Offset;
  samplingTime = millis();
  }
  // printing the measured values to Serial
  if (millis() - printTime > printInterval && inWater)  //Every time printInterval passes, print a numerical, convert the state of the LED indicator
  {
//  Serial.print("Volt: ");
//  Serial.println(voltage);
  Serial.print("P");
  Serial.println(pHValue, 2);
  //  }
  printTime = millis();
  }
}

// GPS functions
inline void GPSPrintFix() {
  fix  = gps.read();
  // print values
  if (fix.valid.location && inWater) {
  Serial.print("Y");
  Serial.println(fix.latitude(), 6 );
  Serial.print("X");
  Serial.println(fix.longitude(), 6 );
  }
}


// pH functions
double averagearray(int* arr, int number) {
  int i;
  int max, min;
  double avg;
  long amount = 0;
  if (number <= 0) {
  //   Serial.println("Error number for the array to averaging!/n");
  return 0;
  }
  if (number < 5) { //less than 5, calculated directly statistics
  for (i = 0; i < number; i++) {
    amount += arr[i];
  }
  avg = amount / number;
  return avg;
  } else {
  if (arr[0] < arr[1]) {
    min = arr[0]; max = arr[1];
  }
  else {
    min = arr[1]; max = arr[0];
  }
  for (i = 2; i < number; i++) {
    if (arr[i] < min) {
      amount += min;    //arr<min
      min = arr[i];
    } else {
      if (arr[i] > max) {
        amount += max;  //arr>max
        max = arr[i];
      } else {
        amount += arr[i]; //min<=arr<=max
      }
    }//if
  }//for
  avg = (double)amount / (number - 2);
  }//if
  return avg;
}
